//
// host.c
// Infektsimulation
//
// Created on 20.08.13.
// Alexander Droste & Florian Flachsenberg
//


#include <stdlib.h>
#include <stdio.h>

#include "agent.h"
#include "host.h"
#include "datatypes.h"
#include "world.h"
#include <gsl_rng.h>
#include <gsl_randist.h>
#include <limits.h>
#include "mpi_datatypes.h"

//accessing global vars with extern
extern World world;
extern gsl_rng *rng;

static void host_find_initial_pos(Host *host);

//get random values - man pages
//https://www.gnu.org/software/gsl/manual/html_node/Sampling-from-a-random-number-generator.html


Host *host_create()
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
	Host *new_host = calloc(1, sizeof(Host));
	new_host->state = HEALTHY;
	new_host->field = NULL;

    //hosts are placed in their own fields
	//find free position for host
    host_find_initial_pos(new_host);
	//minimum of 24 + random[0, 48]
	new_host->disappear_duration = 24 + (rnd_int(rng) % 49);
    
	//immunity list
    new_host->immunities = 0L;

	return new_host;
}

void host_delete(Host *host)
{    
	/* list_delete_element is called in world_remove_host_from_field */
	free(host);
	host = NULL;
}

/**
 *  Find a position for a host during creation process.
 *  Additional function to prevent the possibility of a deadlock
 *  which could happen in host_move during creation process. 
 *  If a host is placed on a reserved field which is surrounded 
 *  by reserved fields there's no way to find a free position.
 *  If not all positions in world are reserved, host_find_initial_pos
 *  will find a free one. Despite, this removes the check for 
 *  host->field != NULL in host_move.
 *
 *  @param host host to find position for
 */
static void host_find_initial_pos(Host *host)
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    unsigned col_rank = rank % world.process_matrix_dimensions.cols;
    unsigned row_rank = rank / world.process_matrix_dimensions.cols;
    unsigned x_offset = world.width_process_mid*col_rank;
    unsigned y_offset = world.height_process_mid*row_rank;
    
	bool keep_searching = true;
	while (keep_searching) {
        host->pos.x = rnd_int(rng) % world.width_process + x_offset;
        host->pos.y = rnd_int(rng) % world.height_process + y_offset;
        
		//if another host is on the same position, keep searching
		keep_searching = world_host_position_collides(host);
	}
}

void host_move(Host *host)
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
	bool keep_searching = true;
    
	unsigned old_x_pos = host->pos.x, old_y_pos = host->pos.y;
	very_short_int rnd_dir_x, rnd_dir_y;
    
	while (keep_searching) {
		//possible values rand%3={0,1,2} -> -1 ={-1, 0, 1}
		rnd_dir_x = (rnd_int(rng) % 3) - 1;
		rnd_dir_y = (rnd_int(rng) % 3) - 1;
        
		host->pos.x = (old_x_pos + rnd_dir_x + world.size.width) % world.size.width;
		host->pos.y = (old_y_pos + rnd_dir_y + world.size.height) % world.size.height;
        
        
        unsigned target_rank = world_get_process_for_host(host);
        
        //if needed send host to other process
        if (rank != target_rank) {
            ++world.hosts_send_count;
            MPI_Request req;
            world_remove_host_from_field(host);
            MPI_Isend(host, 1, mpi_host, target_rank, tag_host, MPI_COMM_WORLD, &req);
            MPI_Wait(&req, MPI_STATUS_IGNORE);
            --world.residents;
            switch (host->state)
            {
              case IMMUNE:
                world.immune_total--;
                world.healthy_current--;
                break;
              case HEALTHY:
                world.healthy_current--;
                break;
              case INFECTED:
                world.infected_total--;
                world.infected_current--;
                break;
              case ILL:
                world.ill_total--;
                world.ill_current--;
                break;
              default:
                ;

            }
            free(host);
//            printf("\nsend from %i => to %i", rank, target_rank);
            //receiving process checks for position collision
            return;
        }
        
		//if another host is on the same position, keep searching
		keep_searching = world_host_position_collides(host);
	}
    
    //host stays in this process but has to switch fields
    Field *correct_field = world_get_field_for_host(host);
    if (host->field != correct_field) {
        world_transfer_host_to_field(host, correct_field);
    }
    
}

void host_start_disease(Host *host)
{
	host->state = ILL;
	world.ill_total++;
    world.ill_current++;
    world.infected_current--;
    int illness_time = host->end_of_disease - host->outbreak_time;
    if (illness_time > world.max_illness) {
        world.max_illness = illness_time;
    }
    else if (illness_time < world.min_illness) {
        world.min_illness = illness_time;
    }
    
    world.sum_illness += (double)illness_time;
}

void host_end_disease(Host *host)
{
    if (host->state == ILL) {
      world.ill_current--;
    }
    else if (host->state == INFECTED) {
      world.infected_current--;
    }
	if (host->will_die) {
		host->state = DEAD;
		world.died_total++;
                world.infectious_dead_current++;
	}
	else if (host->will_immune) {
		host->state = IMMUNE;
        //add immunnity factor
        host->immunities |= host->agent.immunogenicity_factor;
		world.immune_total++;
	}
	else {
		host->state = HEALTHY;
	}
	if (host->state == IMMUNE || host->state == HEALTHY) {
		list_transfer_element(host->field->healthy, host->field->infected, host->list_element);
        host->field->nof_healthy++;
        host->field->nof_infected--;
        world.healthy_current++;
        //is not allocated with malloc -> no deletion process needed
//		host_delete_agent(host);
	}
}


void host_process(Host *host, unsigned short inc_time) {
	if (host->state == DEAD) {
		//after hosts die they stay for the time of their disappear_duration
		host->disappear_duration -= 1;
		//if the duration is over they get removed completely
		if (host->disappear_duration == 0) {
			world_remove_host_from_field(host);
			host_delete(host);
                        world.infectious_dead_current--;
		}
	}
    /* if not dead */
	else {
        /* process infection if host is not healthy */
		if (host->state != HEALTHY && host->state != IMMUNE) {
            /* at outbreak time host infection will turn into disease or not */
			if (inc_time == host->outbreak_time && host->will_ill) {
                /* start disease or skip it*/
                host_start_disease(host);
			}
            /* if disease time is over => end disease */
			else if (inc_time == host->end_of_disease) {
                host_end_disease(host);
			}
		}
	}
}

/**
 * Infect procedure for start of simulation
 *
 * @param host host to get infected
 * @param inc_agent agent infecting the host
 */
void host_infect_sim_start(Host *host, Agent inc_agent)
{
    host_infect(host, inc_agent, 0);
    if (host->state == INFECTED) {
        list_transfer_element(host->field->infected, host->field->healthy, host->list_element);
        //host was added to healthy before
        //correct to inc/decr both counters
        host->field->nof_infected++;
        host->field->nof_healthy--;

        world.infected_total++;
        world.infected_current++;
        world.healthy_current--;
    }
}

/**
 * In case of a infection this function is called.
 *
 * @param host host to get infected
 * @param inc_agent agent infecting the host
 * @param inc_time time point to capture start of infection
 */
void host_infect(Host *host, Agent inc_agent, unsigned inc_time)
{
    bool is_contained = ((host->immunities & inc_agent.immunogenicity_factor)
                           == inc_agent.immunogenicity_factor);

    if (is_contained) {return;}
    
	/* copy agent from other host */
    host->agent = inc_agent;
	host->infection_time = inc_time;
	host->state = INFECTED;
    if (rnd_percent(rng) < host->agent.mutation_prob) {
        agent_mutate(&host->agent);
    }
    
	//immunogenicity checked against random value
	if (rnd_percent(rng) < inc_agent.immunogenicity) {
		host->will_immune = 1;
	}
	else {
		host->will_immune = 0;
	}
    
	//outbreak time gets calculated
	int diff = inc_agent.incubation_max_dur - inc_agent.incubation_mean_dur;
	int time_variation = (rnd_int(rng) % (2 * diff + 1)) - diff;
    int incubation_time = inc_agent.incubation_mean_dur + time_variation;
	host->outbreak_time = inc_time + incubation_time;
    if (incubation_time > world.max_incubation) {
        world.max_incubation = incubation_time;
    }
    else if (incubation_time < world.min_incubation) {
        world.min_incubation = incubation_time;
    }

    world.sum_incubation += (double)incubation_time;


	//outbreak probability checked against random value
	if (rnd_percent(rng) < inc_agent.outbreak_prob) {
		host->will_ill = 1;
        
		//mortality is checked against random value
		if (rnd_percent(rng) < inc_agent.mortality) {
			host->will_die = 1;
		}
		else {
			host->will_die = 0;
		}
        
		//disease endpoint gets calculated        
		diff = inc_agent.illness_max_dur - inc_agent.illness_mean_dur;
		int time_variation_2 = (rnd_int(rng) % (2 * diff + 1)) - diff;
        //disease time calculated beginning with outbreak_time!!
		host->end_of_disease = host->outbreak_time + inc_agent.illness_mean_dur + time_variation_2;

	}
	else {
		host->will_ill = 0;
		host->will_die = 0;
        host->end_of_disease = host->outbreak_time;
	}
}


static void host_infection_process_for_field(Host* host_to_infect, Field *field, unsigned short time)
{
	Host* neigh, *other_host_in_range;
    Host* neigh_array[field->nof_infected];
   
    //put hosts into array
    List_element *el = list_first(field->infected);
    int j = 0;
	while (el) {
        neigh_array[j] = list_element_get_value(el);
        el = list_element_next(el);
        ++j;
    }
   
    //shuffle hosts randomly
    if (field->nof_infected > 1) {
        gsl_ran_shuffle(rng, neigh_array, field->nof_infected, sizeof(Host*));
    }
    
    for (int i = 0; i < field->nof_infected; ++i) {
        neigh = neigh_array[i];
		// list contains only infected and ill host neighbours
		// => are infectious if ill or agent is infectious during incubation
		if ((neigh->agent.infectious_incubation && neigh->state == INFECTED && time > neigh->infection_time) ||
		    (neigh->state == ILL && ((!neigh->agent.infectious_incubation && time > neigh->outbreak_time) ||
              (neigh->agent.infectious_incubation && time > neigh->infection_time))) || neigh->state == DEAD)
              {
                  unsigned dist;
                  other_host_in_range = neigh;
                  dist = sq_euclidean(host_to_infect->pos.x, host_to_infect->pos.y,
                                      other_host_in_range->pos.x, other_host_in_range->pos.y);
                  
                  if (dist < world.radius_air_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent.infection_prob.air) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_oral_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent.infection_prob.oral) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_skin_sq) {
                      if (rnd_percent(rng) < other_host_in_range->agent.infection_prob.skin) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_xxx_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent.infection_prob.xxx) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
              }
	}
}

/**
 * Interacts with all other host in range.
 *
 *
 * @param host host to interact with others
 * @param time time-point of interaction
 */
void host_prepare_infection_process(Host *host, unsigned short time)
{
	Coordinates field_coord_rel = world_get_rel_coord_for_host(host);
    Coordinates field_coord_abs = world_get_abs_for_rel_coord(
      field_coord_rel.x, field_coord_rel.y);
	Position hostPos = world_get_pos_within_field(host);

	unsigned field_max_x_pos = host->field->size.width - 1;
	unsigned field_max_y_pos = host->field->size.height - 1;
    
	//squared euclidean distances to each corner
	unsigned left = sq_euclidean(0, hostPos.y, hostPos.x, hostPos.y);
	unsigned topleft = sq_euclidean(0, 0, hostPos.x, hostPos.y);
	unsigned top = sq_euclidean(hostPos.x, 0, hostPos.x, hostPos.y);
	unsigned topright = sq_euclidean(field_max_x_pos, 0, hostPos.x, hostPos.y);
	unsigned right = sq_euclidean(field_max_x_pos, hostPos.y, hostPos.x, hostPos.y);
	unsigned bottomright = sq_euclidean(field_max_x_pos, field_max_y_pos, hostPos.x, hostPos.y);
	unsigned bottom = sq_euclidean(hostPos.x, field_max_y_pos, hostPos.x, hostPos.y);
	unsigned bottomleft = sq_euclidean(0, field_max_y_pos, hostPos.x, hostPos.y);
  
    //host can have max. 3 neighbourfields + his own = 4
    HostForInfection host_for_infection;
    host_for_infection.host = *host;
    host_for_infection.phost = host;
    //fields processed from max index to 0 (pfield at the end of this process)
    //then no max index var is needed 
    host_for_infection.curr_field_coord = 0;
    
  	//unsigned matrix_cols = world.field_matrix->n * world.process_matrix_dimensions.cols;
  	//unsigned matrix_rows = world.field_matrix->m * world.process_matrix_dimensions.rows;
	  unsigned matrix_cols = world.size.width/world.full_size_one_field.width;
	  unsigned matrix_rows = world.size.height/world.full_size_one_field.height;
    
    host_for_infection.field_coord[host_for_infection.curr_field_coord++] = field_coord_abs;
    
    Coordinates next_field_coord;
    
	if (left < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x - 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = field_coord_abs.y;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (topleft < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x - 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = (field_coord_abs.y - 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (top < world.maxradius_sq) {
		next_field_coord.x = field_coord_abs.x;
        next_field_coord.y = (field_coord_abs.y - 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (topright < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x + 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = (field_coord_abs.y - 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (right < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x + 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = field_coord_abs.y;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (bottomright < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x + 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = (field_coord_abs.y + 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (bottom < world.maxradius_sq) {
		next_field_coord.x = field_coord_abs.x;
        next_field_coord.y = (field_coord_abs.y + 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
	if (bottomleft < world.maxradius_sq) {
		next_field_coord.x = (field_coord_abs.x - 1 + matrix_cols) % matrix_cols;
        next_field_coord.y = (field_coord_abs.y + 1 + matrix_rows) % matrix_rows;
        host_for_infection.field_coord[host_for_infection.curr_field_coord++] = next_field_coord;
	}
    //curr_field_coord ends with max-index +1
    //& is further used to describe the array length
    gsl_ran_shuffle(rng, host_for_infection.field_coord, host_for_infection.curr_field_coord, sizeof(Coordinates));
    host_for_infection_process(&host_for_infection, time);
}

void host_for_infection_process(HostForInfection* host_for_infection, unsigned time)
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    Coordinates local;
    int new_rank;
    //dont end with zero (UNSIGNED)
    while (host_for_infection->curr_field_coord > 0) {
 
        new_rank = world_get_process_for_absolute_coord(
            host_for_infection->field_coord[host_for_infection->curr_field_coord-1].x,
            host_for_infection->field_coord[host_for_infection->curr_field_coord-1].y);
        
        //if we stay => perform infection HERE
        if (rank == new_rank) {
            local = world_get_rel_for_abs_coord(
                  host_for_infection->field_coord[host_for_infection->curr_field_coord-1].x,
                  host_for_infection->field_coord[host_for_infection->curr_field_coord-1].y);
            
            host_infection_process_for_field(&(host_for_infection->host),
            get_field_from_matrix_for_coord(local.x, local.y), time);
            --host_for_infection->curr_field_coord;
            if (host_for_infection->host.state == INFECTED) {break;}
        }
        
        else {
            //printf("\n host send from %i to %i", rank, new_rank);
            ++world.infects_send_count;
            MPI_Request req;
            MPI_Isend(host_for_infection, 1, mpi_hostForInfection, new_rank, tag_infect, MPI_COMM_WORLD, &req);
            // wait, otherwise we may leave the scope and the infection struct is gone
            MPI_Wait(&req, MPI_STATUS_IGNORE);
            --host_for_infection->curr_field_coord;
            return;
        }
    }

    /* the infect-chain has ended
       during the loop break might have been called because of infection
       
       now there are two possibilities: 
         we are already in the "home process"
       => just copy the host from the struct
         we are in another process
       => send it back to the home
     
       if the host is not infected, there is no need for further actions
       because the host has not been changed and needs no notification
    */
    if (host_for_infection->host.state == INFECTED)
    {
       new_rank = world_get_process_for_host(&(host_for_infection->host));  
       if (rank == new_rank) {
           //assign updated values
            *(host_for_infection->phost) = host_for_infection->host;
           static Host* host_to_transfer;
           host_to_transfer = host_for_infection->phost;
           //transfer host to list of infected
           list_transfer_element(host_to_transfer->field->infected,
                                 host_to_transfer->field->healthy,
                                 host_to_transfer->list_element);
           host_to_transfer->field->nof_infected++;
           host_to_transfer->field->nof_healthy--;
           world.infected_current++;
           world.infected_total++;
           world.healthy_current--;
        }
        else {
            ++world.infects_send_count;
            MPI_Request req;
            MPI_Isend(host_for_infection, 1, mpi_hostForInfection, new_rank, tag_infect, MPI_COMM_WORLD, &req);
            // wait, otherwise we may leave the scope and the infection struct is gone
            MPI_Wait(&req, MPI_STATUS_IGNORE);
        }
    }
}
 
static void combine_infectious_lists(List *dest, List *src, unsigned short time) {
	List_element *el = list_first(src);
	Host *neigh;
    
	while (el) {
		neigh = list_element_get_value(el);
		// list contains only infected and ill host neighbours
		// => are infectious if ill or agent is infectious during incubation
		if ((neigh->agent.infectious_incubation &&
		     neigh->state == INFECTED &&
             time > neigh->infection_time) ||
            
		    (neigh->state == ILL &&
             ((!neigh->agent.infectious_incubation &&
               time > neigh->outbreak_time) ||
              (neigh->agent.infectious_incubation &&
               time > neigh->infection_time)))) {
                  list_append(dest, list_element_get_value(el));
              }
		el = list_element_next(el);
	}
}

