//
//  main.c
//  Infektsimulation
//
//  Created on 20.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <limits.h>
#include "host.h"
#include "agent.h"
#include "world.h"
#include <gsl_rng.h> //gsl - high quality random number generators
#include "mpi_datatypes.h"
#include <math.h>

static void print_data_sizes();
static void print_rand_number_setup();
static void print_calculated_values(Agent new_agent);
static void print_calculated_values_mpi(Agent new_agent);

gsl_rng* rng; //rng = random number generator
extern World world;

int main(int argc, char** argv)
{
    MPI_Init(&argc, &argv);
    //size = number of processes
    int size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (argc < 6) {
        if (rank == 0) {
            fprintf(stderr,"\n!wrong argument count\ncorrect usage:\n%s world_size number_hosts percentage_infected\n"\
                    "number_of_steps agent stop-flag\nworld_size: length of world axis\n"\
                    "number_hosts: initial number of hosts\n"\
                    "percenta_infected: percentage of hosts initially infected\n"\
                    "number_of_steps: number of time steps of simulation\n"\
                    "agent: agenty type (0: Influenza, 1: Ebola, 3: Herpes, 4: Smallpox)\n"\
                    "stop-flag: 0, if no stopping after all infectious hosts are gone\n", argv[0]);
        }
        return 1;
    }
    unsigned world_size, number_of_hosts;
    float percentage_infected;
    unsigned time_steps, agent_code, stop_if_finished;

    if (sscanf(argv[1], "%u", &world_size) != 1) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid world size!\n", argv[1]);
        }
        return 1;
    }

    if (sscanf(argv[2], "%u", &number_of_hosts) != 1) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid number of hosts!\n", argv[2]);
        }
        return 1;
    }

    if (sscanf(argv[3], "%f", &percentage_infected) != 1 || 
        percentage_infected < 0.0 || percentage_infected > 100.0) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid infection percentage!\n", argv[3]);
        }
        return 1;
    }

    if (sscanf(argv[4], "%u", &time_steps) != 1) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid number of time steps!\n", argv[4]);
        }
        return 1;
    }

    if (sscanf(argv[5], "%u", &agent_code) != 1 || agent_code > 5) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid agent code!\n", argv[5]);
        }
        return 1;
    }

    if (sscanf(argv[6], "%u", &stop_if_finished) != 1 || stop_if_finished > 1) {
        if (rank == 0) {
            fprintf(stderr,"error: %s is no valid stop-flag (0 or 1)!\n", argv[6]);
        }
        return 1;
    }


    initialize_mpi_datatypes();
    
    rng = gsl_rng_alloc(gsl_rng_knuthran);//each process has his own rng in mpi
    gsl_rng_set(rng, rank); //use different initial value/seed to get different outcome
    //http://fossies.org/dox/gsl-1.16/knuthran_8c_source.html

    int i;
    Agent new_agent = agent_new_stereotype(agent_code);
    if (!agent_is_valid(new_agent)) {
        if (rank == 0) {
            fprintf(stderr, "specified agent is not valid!\n");
        }
        return 1;
    }
  
    if (rank == 0) {
        printf("number of hosts %i\n", number_of_hosts);
        printf("world size %i\n", world_size);
        printf("time steps %i\n", time_steps);
        printf("start percentage infected %.1f\n", percentage_infected);
    }
    //create world with size (axes length^2), agent
    world_create_for_agent(world_size, new_agent);

    //create i hosts and add them to the world
    for (i = 0; i < number_of_hosts / size; ++i) {
        Host* new_host = host_create();
        world_add_host_to_field(new_host);
        //printf("%u, %u\n", new_host->pos.x, new_host->pos.y);
        ++world.residents;
        world.healthy_current++;
        //infect x percent of hosts initially
        //given by percentage_infected
        if (i < ceil((number_of_hosts / size) * (percentage_infected/100))) {
            host_infect_sim_start(new_host, new_agent);
        }
    }
    //simulate x time steps
    world_simulate(time_steps, stop_if_finished);
    
    print_calculated_values_mpi(new_agent);
//    print_data_sizes();
//    print_rand_number_setup()
    
    world_destroy();
    MPI_Finalize();
    return 0;
}

static void print_calculated_values(Agent new_agent)
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    printf("\ncalculated_values from rank:%i\n",rank);

    printf("back calculated values (CAUTION, only valid without mutations)\n");
    printf("total infected: %u\n", world.infected_total);
    printf("total ill: %u (expected from total infected and agent outbreak prob: %.1f)\n",
           world.ill_total, world.infected_total*new_agent.outbreak_prob/100.0);
    printf("total died: %u (expected from total ill and agent mortality: %.1f)\n",
           world.died_total, world.ill_total*new_agent.mortality/100.0);
    printf("total immune: %u (expected from total infected and immunogenic prob: %.1f)\n",
           world.immune_total, world.infected_total*new_agent.immunogenicity/100.0);
    printf("incubation time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n",
           world.sum_incubation/world.infected_total, world.min_incubation, world.max_incubation,
           new_agent.incubation_mean_dur,
           2*new_agent.incubation_mean_dur - new_agent.incubation_max_dur,
           new_agent.incubation_max_dur);
    printf("illness time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n",
           world.sum_illness/world.ill_total, world.min_illness, world.max_illness,
           new_agent.illness_mean_dur,
           2*new_agent.illness_mean_dur - new_agent.illness_max_dur,
           new_agent.illness_max_dur);
    printf("\n");

}

static void print_calculated_values_mpi(Agent new_agent)
{
    int size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    unsigned infected_total_sum;
    MPI_Reduce(&world.infected_total, &infected_total_sum, 1,
               MPI_UNSIGNED, MPI_SUM, 0, MPI_COMM_WORLD);
    
    unsigned ill_total_sum;
    MPI_Reduce(&world.ill_total, &ill_total_sum, 1,
               MPI_UNSIGNED, MPI_SUM, 0, MPI_COMM_WORLD);
    
    unsigned died_total_sum;
    MPI_Reduce(&world.died_total, &died_total_sum, 1,
               MPI_UNSIGNED, MPI_SUM, 0, MPI_COMM_WORLD);
    
    unsigned immune_total_sum;
    MPI_Reduce(&world.immune_total, &immune_total_sum, 1,
               MPI_UNSIGNED, MPI_SUM, 0, MPI_COMM_WORLD);
    
    
    
    double sum_incubation, mean_incubation;
    MPI_Reduce(&world.sum_incubation, &sum_incubation, 1,
               MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

    mean_incubation = sum_incubation / infected_total_sum;
    
    unsigned min_incubation;
    MPI_Reduce(&world.min_incubation, &min_incubation, 1,
               MPI_UNSIGNED, MPI_MIN, 0, MPI_COMM_WORLD);
    
    unsigned max_incubation;
    MPI_Reduce(&world.max_incubation, &max_incubation, 1,
               MPI_UNSIGNED, MPI_MAX, 0, MPI_COMM_WORLD);
    
    
    double sum_illness, mean_illness;
    MPI_Reduce(&world.sum_illness, &sum_illness, 1,
               MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

    mean_illness = sum_illness / ill_total_sum;

    unsigned min_illness;
    MPI_Reduce(&world.min_illness, &min_illness, 1,
               MPI_UNSIGNED, MPI_MIN, 0, MPI_COMM_WORLD);
    
    unsigned max_illness;
    MPI_Reduce(&world.max_illness, &max_illness, 1,
               MPI_UNSIGNED, MPI_MAX, 0, MPI_COMM_WORLD);
    
    
    if (rank == 0) {
        printf("\nback calculated values (CAUTION, only valid without mutations)");
        printf("\n–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––\n");
        printf("total infected: %u\n\n", infected_total_sum);
        printf("total ill: %u (expected from total infected and agent outbreak prob: %.1f)\n\n",
               ill_total_sum, infected_total_sum*new_agent.outbreak_prob/100.0);
        printf("total died: %u (expected from total ill and agent mortality: %.1f)\n\n",
               died_total_sum, ill_total_sum*new_agent.mortality/100.0);
        printf("total immune: %u (expected from total infected and immunogenic prob: %.1f)\n\n",
               immune_total_sum, infected_total_sum*new_agent.immunogenicity/100.0
               *(100.0-new_agent.outbreak_prob)/100.0
               + ill_total_sum*(100.0-new_agent.mortality)/100.0*new_agent.immunogenicity/100.0);
        
        printf("incubation time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n\n",
               //actual
               mean_incubation,
               min_incubation,
               max_incubation,
               //expected
               new_agent.incubation_mean_dur,
               2*new_agent.incubation_mean_dur - new_agent.incubation_max_dur,
               new_agent.incubation_max_dur);
        
        printf("illness time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n",
               //actual
               mean_illness,
               min_illness,
               max_illness,
               //expected
               new_agent.illness_mean_dur,
               2*new_agent.illness_mean_dur - new_agent.illness_max_dur,
               new_agent.illness_max_dur);
        printf("–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––\n");
    }
}


static void print_data_sizes()
{
    int size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    
    if (rank == 0) {
        printf("char: %lu byte\n",sizeof(char));
        printf("bool: %lu byte\n",sizeof(bool));
        printf("short: %lu bytes\n",sizeof(short));
        printf("int: %lu bytes\n",sizeof(int));
        printf("long: %lu bytes\n",sizeof(long));
        printf("float: %lu bytes\n",sizeof(float));
        printf("double: %lu bytes\n",sizeof(double));
        
        printf("unsigned int max value: %u\n",UINT_MAX);
        printf("unsigned long max value: %lu\n",ULONG_MAX);
        printf("random max value: %d", RAND_MAX);

        printf("\n");
    }
}

static void print_rand_number_setup()
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    if (rank == 0) {
        printf("random gener. min %lu\n", gsl_rng_min(rng));
        printf("random gener. max %lu\n", gsl_rng_max(rng));
        //returns integer in range [0 - rng_max]
        printf("get rand int_1: %li int_2: %li\n", rnd_int(rng), rnd_int(rng));
        //returns double in range [0 - 1), 1 is excluded
        printf("get rand float_1: %f float_2: %f\n\n", gsl_rng_uniform(rng), gsl_rng_uniform(rng));
    }
}
