//
//  world.c
//  Infektsimulation
//
//  Created on 21.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include "world.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include "mpi_datatypes.h"

#ifdef DRAW
#define MAGICKCORE_QUANTUM_DEPTH 16
#define MAGICKCORE_HDRI_ENABLE 0
#include <magick_wand.h>
#endif

#define INVALID_SIZE_CONSTELLATION 667
#define INVALID_PROCESS_FIELD_CONSTELLATION 668
#define INVALID_PROCESS_HEIGHT_OR_WIDTH 669
#define INVALID_MATRIX_DIMENSIONS 670

//global world var can get accessed through the "extern" modifier in other files
World world;
//––––––––––––––––private functions––––––––––––––––––––––––––––––
static MatrixDimensions matrix_rows_cols_for_size(unsigned n);
static void world_move_healthy_in_field(unsigned x, unsigned y, DynArray* neigh_processes);
static void world_move_infected_in_field(unsigned x, unsigned y, DynArray* neigh_processes);
static void world_move_both_types_in_field(unsigned x, unsigned y, DynArray* neigh_processes);
static void world_process_healthy_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time);
static void world_process_infected_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time);
static void world_process_both_types_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time);
static int world_log_timestep_stats(unsigned short curr_time_step);
#ifdef DRAW
static void world_draw(unsigned timestep);
#endif
//–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

/**
 *  Calculate matrix dimensions for a given integer.
 *  The factors obtained by prime factorization are
 *  distributed as even as possible between rows/cols.
 *
 *  @param n number to calc matrix dim. for
 *
 *  @return calculated matrix dimensions
 */
static MatrixDimensions matrix_rows_cols_for_size(unsigned n) {
    
    List* l = list_new();
    unsigned orig_value = n;
    double sqrt_n = sqrt (n);
    unsigned mul = 1;
    
    //find prime factors
    if (n >= 2) {
        int i = 2;
        while (i <= sqrt_n) {
            while (n % i == 0) {
                unsigned * s = calloc(1, sizeof(unsigned));
                *s = i;
                list_append(l, s);
                mul *= i;
                //printf("%i",*s);
                n /= i;
            }
            ++i;
        }
        if (mul < orig_value) {
            unsigned * s = calloc(1, sizeof(unsigned));
            *s = (orig_value / mul);
            list_append(l, s);
        }
    }
    
    MatrixDimensions md = {.rows=1, .cols=1};
    
    /* distribute factors, so that
       col-row count are as even as possible.
       start with biggest prime factors. */
    List_element* le = list_last(l);
    while (le) {
        //if cols smaller, continue with cols
        if (md.rows > md.cols){
            md.cols *= (*(unsigned *)le->data);
        }
        //if rows smaller, continue with rows
        else if (md.cols > md.rows){
            md.rows *= (*(unsigned *)le->data);
        }
        //if equally big continue with cols
        else{
            md.cols *= (*(unsigned *)le->data);
        }
        
        le = list_element_prev(le);
    }
    
    int rank; MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    if (rank == 0) {
        printf("(%icols x %irows)\n", md.cols, md.rows);
    }
    list_delete(l);
    
    #ifdef DEBUG
        assert(md.rows * md.cols == orig_value);
    #endif
    
    return md;
}


/**
 *  initialises, creates the world, defines world size, fields size
 *
 *  @param world_axes_length defines width and height of the world
 *  @param field_axes_length defines width and height for all fields
 */
void world_create(unsigned int world_axes_length, unsigned int field_axes_length)
{
    //use of static to construct singleton
    static bool  world_is_created = false;
    if (!world_is_created) {
        
        int size, rank;
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        MPI_Comm_size(MPI_COMM_WORLD, &size);
     
        if (rank == 0) {printf("\nprocess dim.: ");}
        //assign how the processes are arranged
        world.process_matrix_dimensions = matrix_rows_cols_for_size(size);
        
        world.size.width = world_axes_length;
        world.size.height = world_axes_length;
        world.full_size_one_field.width = field_axes_length;
        world.full_size_one_field.height = field_axes_length;

        
        //initial values for statistic
        world.immune_total = 0;
        world.infected_total = 0;
        world.died_total = 0;
        world.ill_total = 0;
        world.ill_current = 0;
        world.infected_current = 0;
        world.healthy_current = 0;
        world.infectious_dead_current = 0;
        world.hosts_send_count = 0;
        world.hosts_received_count = 0;
        world.residents = 0;
        
        world.min_incubation = world.min_illness = USHRT_MAX;
        world.max_incubation = world.max_illness = 0;
        world.sum_incubation = world.sum_illness = 0.0;

        world.radius_air_sq = radius_air*radius_air;
        world.radius_skin_sq = radius_skin*radius_skin;
        world.radius_oral_sq = radius_oral*radius_oral;
        world.radius_xxx_sq = radius_xxx*radius_xxx;

        //find max radius of transmission types
        world.maxradius_sq = max(max(world.radius_air_sq, world.radius_oral_sq),
                                 max(world.radius_skin_sq, world.radius_xxx_sq));


       
        double number_of_fields_total = ((double)world_axes_length*world_axes_length) /
                                    ((double)field_axes_length*field_axes_length);
       
        
        unsigned rank_col = rank % world.process_matrix_dimensions.cols;
        unsigned rank_row = rank / world.process_matrix_dimensions.cols;

        double num_block_field_width = world.size.width/field_axes_length;
        double num_block_field_height = world.size.height/field_axes_length;


        if (num_block_field_width > floor(num_block_field_width)) {
            printf("\nerror: number of fields in process cannot be distributed on block-fields!\n\n");
            exit(INVALID_PROCESS_FIELD_CONSTELLATION);
        }
    
        if (num_block_field_height > floor(num_block_field_height)) {
            printf("\nerror: number of fields in process cannot be distributed on block-fields!\n\n");
            exit(INVALID_PROCESS_FIELD_CONSTELLATION);
        }
       
//        if (rank == 0) {
//          printf("\n%.3f x %.3f\n", num_block_field_height, num_block_field_width);
//        }
        unsigned number_of_cols_this_process = ceil(num_block_field_width/world.process_matrix_dimensions.cols);
        world.width_process_mid = number_of_cols_this_process * field_axes_length;
        world.number_of_cols_mid = number_of_cols_this_process;

        if (rank_col == world.process_matrix_dimensions.cols-1) {
            
            number_of_cols_this_process = num_block_field_width 
              - ((world.process_matrix_dimensions.cols-1) * number_of_cols_this_process);
        }

        world.width_process = number_of_cols_this_process * field_axes_length;
        world.number_of_cols = number_of_cols_this_process;


        unsigned number_of_rows_this_process = ceil(num_block_field_height/world.process_matrix_dimensions.rows);
        world.height_process_mid = number_of_rows_this_process * field_axes_length;
        world.number_of_rows_mid = number_of_rows_this_process;

        if (rank_row == world.process_matrix_dimensions.rows-1) {
            //printf("\n%i\n", rank);
            number_of_rows_this_process = num_block_field_height
              - ((world.process_matrix_dimensions.rows-1) * number_of_rows_this_process);
        }

        world.height_process = number_of_rows_this_process * field_axes_length;
        world.number_of_rows = number_of_rows_this_process;

        double number_of_fields_this_process = world.height_process * world.width_process;

        if (size > number_of_fields_total){
            if(rank == 0){printf("\nerror: more processes than fields\n\n");}
            exit(INVALID_PROCESS_FIELD_CONSTELLATION);
        }
        
        if (number_of_fields_this_process < 100){
            if(rank == 0){printf("\nerror: use less processes => fields per process >= 100\n\n");}
            exit(INVALID_PROCESS_FIELD_CONSTELLATION);
        }
        
        //check if world size can be splitted into fields
        if (number_of_fields_total > floor(number_of_fields_total)) {
            if(rank == 0){printf("\nerror: world_size cannot be splitted into fields\n\n");}
            exit(INVALID_SIZE_CONSTELLATION);
        }

        
        if (rank == 0) {
            printf("process count: %i                         \
                   \nfields total: %i                         \
                   \nfields per process: %i                   \
                   \nworld surface: %llu                      \
                   \none field surface: %u\n",
                    size,
                   (int)number_of_fields_total,
                   (int)number_of_fields_this_process,
                   (unsigned long long)world.size.width*(unsigned long long)world.size.height,
                   field_axes_length*field_axes_length);
        }

       
        if (rank == 0) {printf("matrix per process rows:%i cols%i\n",
                               number_of_rows_this_process, number_of_cols_this_process);}
        //creates matrix with dimensions needed by !one process
        world.field_matrix = createMatrix(number_of_rows_this_process, number_of_cols_this_process, field_axes_length);
        if (world.height_process_mid != world.height_process
            || world.width_process_mid != world.width_process) {
            printf("\nmatrix rank %i: %i x %i\n", rank, number_of_rows_this_process, number_of_cols_this_process);
        }
        
        
        world_is_created = true;
    }
}

/**
 *  Destroy/remove all fields, lists, hosts, etc..
 *  -> all prev allocated memory
 */
void world_destroy()
{
    List_element* el_host[2];
    Host* host;
    Field* f;
    //remove all hosts from fields, delete them
    for (int x = 0; x < world.field_matrix->n ; ++x) {
        for (int y = 0; y < world.field_matrix->m; ++y) {
            f = get_field_from_matrix_for_coord(x, y);
            el_host[0] = list_first(f->healthy);
            el_host[1] = list_first(f->infected);

            for (int i = 0; i < 2; ++i) {
                while (el_host[i]) {
                    host = list_element_get_value(el_host[i]);
                    el_host[i] = list_element_next(el_host[i]);
                    world_remove_host_from_field(host);
                    host_delete(host);
                }
            }
        }
    }
    //free matrix, lists included
    freeMatrix(world.field_matrix);
}


/**
 *  initialises, creates the world, defines world size, fields size
 *  is optimized for agent
 *  this function optimizes the field size and ignores zero-probabilty-transmissions
 *
 *  @param world_axes_length defines width and height of the world
 *  @param agent specifies Agent the parameters should be optimized for
 *
 */
void world_create_for_agent(unsigned world_axes_length, Agent agent)
{
    unsigned maxr = 0;
    //find max radius of all transmission types
    if (agent.infection_prob.air != 0 && radius_air > maxr) maxr = radius_air;
    if (agent.infection_prob.skin != 0 && radius_skin > maxr) maxr = radius_skin;
    if (agent.infection_prob.oral != 0 && radius_oral > maxr) maxr = radius_oral;
    if (agent.infection_prob.xxx != 0 && radius_xxx > maxr) maxr = radius_xxx;
    
    //maxr * x defines field axis length
    unsigned short field_axis_length_factor = 2;
    world_create(world_axes_length, maxr *  field_axis_length_factor);
    world.maxradius_sq = maxr*maxr;
}

/**
 *  Transfers a host to a another field
 *
 *  @param inc_host host to transfer
 *  @param dest_field field to transfer host to
 */
void world_transfer_host_to_field(Host* inc_host, Field* dest_field)
{
    #ifdef DEBUG
        assert(inc_host->field != NULL);
        assert(dest_field != NULL);
        assert(inc_host->field != dest_field);
    #endif
    
    List* dest_list;
    List* src_list;
    

    if(inc_host->state == HEALTHY || inc_host->state == IMMUNE){
        src_list = inc_host->field->healthy;
        inc_host->field->nof_healthy--;
        dest_list = dest_field->healthy;
        dest_field->nof_healthy++;
    }
    else{
        src_list = inc_host->field->infected;
        inc_host->field->nof_infected--;
        dest_list = dest_field->infected;
        dest_field->nof_infected++;
    }
    //destination field must be assigned to hosts attribute
    inc_host->field = dest_field;
    list_transfer_element(dest_list, src_list, inc_host->list_element);
}

/**
 *  Calculate the appropriate field for a host and add it to it
 *
 *  @param inc_host host to add to a field based on his position
 */
void world_add_host_to_field(Host* inc_host)
{
    List* list;
    Field* wf = world_get_field_for_host(inc_host);
    
    //host knows his field
    inc_host->field = wf;
    
    if (inc_host->state == HEALTHY || inc_host->state == IMMUNE) {
        list = wf->healthy;
        wf->nof_healthy++;
    }
    else {
        list = wf->infected;
        wf->nof_infected++;
    }
    /* simplifies deletion -> list_delete_element
       used in world_remove_host_from_field */
    inc_host->list_element = list_append(list, inc_host);    
    
    #ifdef DEBUG
        int rank; MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        assert(rank == world_get_process_for_host(inc_host));
    #endif
}

/**
 *  Remove a host from his field.
 *  Should only be called after host died and
 *  dissappear duration has passed. Must be called before host_delete.
 *
 *  @param inc_host host to remove
 */
void world_remove_host_from_field(Host* inc_host)
{
    List* list;
    //the hosts assigned field must be used
    Field* wf = inc_host->field;
    if (inc_host->state == HEALTHY || inc_host->state == IMMUNE) {
        list = wf->healthy;
        wf->nof_healthy--;
    }
    else {
        list = wf->infected;
        wf->nof_infected--;
    }
    inc_host->field = NULL;
    list_delete_element(list, inc_host->list_element);
}

/**
 *  Find out if another host is on the same position
 *
 *  @param inc_host the host to check the collision for
 *
 *  @return boolean value if there is a collision
 */
bool world_host_position_collides(Host* inc_host)
{

    Field* wf = world_get_field_for_host(inc_host);
    List_element* el_healthy = list_first(wf->healthy);
    List_element* el_infected = list_first(wf->infected);
    #ifdef PRINT_COLLISION
        static unsigned int colli = 0;
    #endif
    
    //initialised with collision set to 0
    bool collision = 0;

    //iterate over all elements in list
    while (el_healthy) {
        Host* host_from_list = el_healthy->data;

        //if the list element is not the given host itself
        if (host_from_list != inc_host) {
            if(host_from_list->pos.x == inc_host->pos.x &&
               host_from_list->pos.y == inc_host->pos.y){
                #ifdef PRINT_COLLISION
                    printf("collision no: %u\n", ++colli);
                #endif
                collision = 1; //if positions are equal
                break; //end while-loop
            }
        }
        el_healthy = list_element_next(el_healthy);

    }
    //if no collision in first loop
    if (!collision) {
        while (el_infected) {
            Host* host_from_list = el_infected->data;
            
            //if the list element is not the given host itself
            if (host_from_list != inc_host) {
                if(host_from_list->pos.x == inc_host->pos.x &&
                   host_from_list->pos.y == inc_host->pos.y){
                    #ifdef PRINT_COLLISION
                        printf("collision no: %u\n", ++colli);
                    #endif
                    collision = 1; //if positions are equal
                    break; //end while-loop
                }
            }
            el_infected = list_element_next(el_infected);
        }
    }
    return collision;
}

/**
*  Get the correct field for a host, based on his position
*  Which field is chosen is calculated in world_get_rel_coord_for_host
*  !Precondition is that host is already in the correct process!
*
*  @param inc_host given host
*
*  @return field (list of hosts in same field)
*/
Field* world_get_field_for_host(Host* inc_host)
{
    Coordinates field_coordinates = world_get_rel_coord_for_host(inc_host);

    Field* wf =
    get_field_from_matrix_for_coord(field_coordinates.x, field_coordinates.y);
    
    #ifdef DEBUG
        assert (wf);
    #endif

    return wf;

}

Coordinates world_get_abs_for_rel_coord(unsigned x, unsigned y)
{
    Coordinates abs;
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    unsigned rank_col = rank % world.process_matrix_dimensions.cols;
    unsigned rank_row = rank / world.process_matrix_dimensions.cols;

    abs.x = rank_col * world.number_of_cols_mid + x;
    abs.y = rank_row * world.number_of_rows_mid + y;

    #ifdef DEBUG
      assert (rank_col < world.process_matrix_dimensions.cols);
      assert (rank_row < world.process_matrix_dimensions.rows);
      assert (abs.x < (world.process_matrix_dimensions.cols-1) * 
          world.number_of_cols_mid
          + world.field_matrix->n);
      assert (abs.y < (world.process_matrix_dimensions.rows-1) * 
          world.number_of_rows_mid
          + world.field_matrix->m);
    #endif

    return abs;
}

unsigned world_get_process_for_absolute_coord(unsigned x, unsigned y)
{
    unsigned rank_row = y / world.number_of_rows_mid;
    unsigned rank_col = x / world.number_of_cols_mid;
    unsigned rank = rank_col + rank_row * world.process_matrix_dimensions.cols;
    
    int size;
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    #ifdef DEBUG
      assert(rank < size);
    #endif

    return rank;
}

Coordinates world_get_rel_for_abs_coord(unsigned x, unsigned y)
{
    Coordinates field;
    field.y = y % world.number_of_rows_mid;
    field.x = x % world.number_of_cols_mid;

    return field;
}

/**
 *  Get the appropriate process for a host, based on his position.
 *  Counting starts with rows, then cols and progresses like:
 *  0 1 2  <- first row, processed by ranks (0,1,2)
 *  3 4 5  <- sec. row, rocessed by ranks (3,4,5) 
 *
 *  @param inc_host host to find the process for.
 *
 *  @return appropriate process
 */
unsigned world_get_process_for_host(Host* inc_host)
{
    int size;
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    
    Coordinates coord;
   
    coord.x = floor(inc_host->pos.x / world.width_process_mid);
    coord.y = floor(inc_host->pos.y / world.height_process_mid);
   
    return (coord.x + (coord.y  * world.process_matrix_dimensions.cols));
}

/**
 *  Get the coordinates for the field the host is positioned in.
 *
 *  @param inc_host host to find out the field coordinates for
 *
 *  @return field coordinates
 */
Coordinates world_get_rel_coord_for_host(Host* inc_host)
{
    Coordinates field_coordinates;
    #if DEBUG
        //positioning uses %world.size->pos. values must be smaller
        assert(!(inc_host->pos.x >= world.size.width) ||
               (inc_host->pos.y >= world.size.height));
    #endif
    
    field_coordinates.x = floor((inc_host->pos.x % world.width_process_mid) / world.full_size_one_field.width);
    field_coordinates.y = floor((inc_host->pos.y % world.height_process_mid) / world.full_size_one_field.height);

    // check if host has max x coord
    if (((double)(inc_host->pos.x % world.width_process_mid)) / (double)world.full_size_one_field.width >
        world.width_process - world.full_size_one_field.width) {
        field_coordinates.x += 1;
    }
    
    #if DEBUG
        assert((field_coordinates.x < world.field_matrix->n) &&
           (field_coordinates.y < world.field_matrix->m));
    #endif
    return field_coordinates;
}

/**
 *  Get the coordinates within a field for a "world"-position.
 *  The coordinates are in range of the general field width/height.
 *
 *  @param x x-position
 *  @param y y-position
 *
 *  @return coordinates within the field
 */
Position world_get_pos_within_field(Host* host)
{
    Position pos;
    pos.x = host->pos.x % host->field->size.width;
    pos.y = host->pos.y % host->field->size.height;

    return pos;
}

/**
 *  Draw positions and state of all hosts for 
 *  the process the function is called in.
 *  To get a complete picture all processes
 *  must run this function.
 *
 *  @param timestep current time step
 */
#ifdef DRAW
static void world_draw(unsigned timestep)
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    #ifdef ALEX
        static const char* path = "/home/droste/projects/Infektsimulation_Parallel/PICT/ts";
    #endif
    #ifdef FLORIAN
        static const char* path ="~";
    #endif
    
    MagickWand* m_wand = NULL;
    PixelWand* c_wand = NULL;
    DrawingWand* d_wand = NULL;

    char filename[256];

    MagickWandGenesis();

    m_wand = NewMagickWand();
    c_wand = NewPixelWand();
    d_wand = NewDrawingWand();
    PixelSetColor(c_wand, "white");
    MagickNewImage(m_wand, world.width_process, world.height_process, c_wand);

    unsigned x,y;
    Field* wf;
    List_element* el;
    Host* host;
    

    for (x = 0; x < world.field_matrix->n; ++x) {
        for (y = 0; y < world.field_matrix->m; ++y) {
            wf = get_field_from_matrix_for_coord(x, y);
            el = list_first(wf->healthy);
            while (el) {
                host = list_element_get_value(el);
                switch (host->state) {
                    case HEALTHY:
                      PixelSetColor(c_wand, "green");
                      break;    
                    case IMMUNE:
                      PixelSetColor(c_wand, "blue");
                      break;
                    default:
                      PixelSetColor(c_wand, "cyan");
                      printf("warning: unknown host state!\n");

                }
                DrawSetFillColor(d_wand, c_wand);
                DrawPoint(d_wand, host->pos.x%world.width_process_mid, host->pos.y%world.height_process_mid);
                el = list_element_next(el);
            }
            el = list_first(wf->infected);
            while (el) {
                host = list_element_get_value(el);
                switch (host->state) {
                    case ILL:
                      PixelSetColor(c_wand, "red");
                      break;
                    case INFECTED:
                      PixelSetColor(c_wand, "yellow");
                      break;
                    case DEAD:
                      PixelSetColor(c_wand, "black");
                      break;
                    default:
                      PixelSetColor(c_wand, "cyan");
                      printf("warning: unknown host state!\n");
                }
                DrawSetFillColor(d_wand, c_wand);
                DrawPoint(d_wand, host->pos.x%world.width_process_mid, host->pos.y%world.height_process_mid);
                el = list_element_next(el);
            }
        }
    }

    memset(filename, '\0', 256);
    sprintf(filename, "%s_%i__rank_%u.png", path, timestep, rank);

    MagickDrawImage(m_wand, d_wand);

    MagickWriteImage(m_wand,filename);
    c_wand = DestroyPixelWand(c_wand);
    m_wand = DestroyMagickWand(m_wand);
    d_wand = DestroyDrawingWand(d_wand);

    MagickWandTerminus();
}
#endif

/**
 *  Asynchronous receiving of hosts.
 *  Is called in world_process_healthy/infected_in_field
 *  to receive hosts who run through a field of this process
 *  during their infection procedure.
 *
 *  @param neigh_processes neighbour processes which are
 *  candidates to send hosts to this process
 *  @param time current time step
 */
void world_polling_process(DynArray* neigh_processes, unsigned time)
{
    static MPI_Request request_host[8] =
    {MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL};
    static HostForInfection recv_buff[8];
    
    int received_message = 0;
    static int check_for_message[8] = {1,1,1,1,1,1,1,1};
    
    //check for all neighbours if they send a host
    for (int i = 0; i < world.number_of_neighbours; ++i) {
        
        if (check_for_message[i]) {
            //this polling works with local host copies
            MPI_Irecv(&recv_buff[i], 1, mpi_hostForInfection, neigh_processes->memory[i],
                      tag_infect, MPI_COMM_WORLD, &request_host[i]);
            check_for_message[i] = 0;
            
        }
    }
    //check if host was received
    static int i;
    MPI_Testany(8, request_host, &i, &received_message, MPI_STATUS_IGNORE);
    
    if (received_message) {
        ++world.infects_received_count;
        
        static HostForInfection received_host_for_infection;
        received_host_for_infection = recv_buff[i];
        request_host[i] = MPI_REQUEST_NULL;
        check_for_message[i] = 1;
        //received_message is set dyn. in MPI_Test
        
        // now let the host for infection be processed
        // if the infection chain is ended, host is finally
        // infected, it is sent back to its home
        host_for_infection_process(&received_host_for_infection, time);
        
    }
}

/**
 *  Asynchronous receiving of hosts.
 *  Is called in world_move_healthy/infected_in_field
 *  to receive hosts who should be placed in this process.
 *
 *  @param neigh_processes neighbour processes which are
 *         candidates to send hosts to this process
 */
void world_polling_move(DynArray* neigh_processes)
{
    static MPI_Request request_host[8] =
    {MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL,MPI_REQUEST_NULL};
    
    static Host* recv_buff[8];
    
    int received_message = 0;
    static int check_for_message[8] = {1,1,1,1,1,1,1,1};
    
    int rank; MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    //check for all neighbours if host sent
    for (int i = 0; i < world.number_of_neighbours; ++i) {
        
        if (check_for_message[i]) {
            recv_buff[i] = calloc(1, sizeof(Host));
            MPI_Irecv(recv_buff[i], 1, mpi_host, neigh_processes->memory[i],
                      tag_host, MPI_COMM_WORLD, &request_host[i]);
            check_for_message[i] = 0;
            
        }
    }
    //check if host was received
    static int i;
    MPI_Testany(8, request_host, &i, &received_message, MPI_STATUS_IGNORE);
    
    if (received_message) {
        ++world.hosts_received_count;
        // printf("\nreceived in %i <= from %i", rank, neigh_processes->memory[i]);
        
        Host* received_host = recv_buff[i];
        recv_buff[i] = NULL;
        world_add_host_to_field(received_host);
        
        request_host[i] = MPI_REQUEST_NULL;
        check_for_message[i] = 1;
        //received_message is set dyn. in MPI_Test
        
        ++world.residents;
        
        switch(received_host->state) 
        {
          case IMMUNE:
            world.immune_total++;
            world.healthy_current++;
            break;
          case HEALTHY:
            world.healthy_current++;
            break;
          case INFECTED:
            world.infected_current++;
            world.infected_total++;
            break;
          case ILL:
            world.ill_current++;
            world.ill_total++;
            break;
          default:
              ;
        }

        //in case received host is not on a free position
        if (world_host_position_collides(received_host)) {
            host_move(received_host);
        }
    }
    
    
}

/**
 *  Move all healthy/immune hosts in a field.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 */
static void world_move_healthy_in_field(unsigned x, unsigned y, DynArray* neigh_processes)
{
    static Field* wf;
    static Host* host;
    static List_element* el_host;
    wf = get_field_from_matrix_for_coord(x, y);
    //move all healthy
    el_host = list_first(wf->healthy);
    while (el_host) {
        world_polling_move(neigh_processes);
        host = list_element_get_value(el_host);
        el_host = list_element_next(el_host);
        host_move(host);//can send hosts to other processes
    }
}

/**
 *  Move all infected/ill hosts in a field.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 */
static void world_move_infected_in_field(unsigned x, unsigned y, DynArray* neigh_processes)
{
    static Field* wf;
    static Host* host;
    static List_element* el_host;
    wf = get_field_from_matrix_for_coord(x, y);
    //move all infected
    el_host = list_first(wf->infected);
    while (el_host) {
        world_polling_move(neigh_processes);
        host = list_element_get_value(el_host);
        el_host = list_element_next(el_host);
        if (host->state != DEAD) {
            host_move(host);//can send hosts to other processes
        }
    }
}

/**
 *  Convenience function which wraps both move type functions.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 */
static void world_move_both_types_in_field(unsigned x, unsigned y, DynArray* neigh_processes)
{
    world_move_healthy_in_field(x, y, neigh_processes);
    world_move_infected_in_field(x, y, neigh_processes);
}

/**
 *  Process all healthy/immune hosts for a field.
 *  The difference after processing is the result 
 *  of one time step evaluated.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 *  @param time            current time step
 */
static void world_process_healthy_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time)
{
    static Field* wf;
    static Host* host;
    static List_element* el_host;
    wf = get_field_from_matrix_for_coord(x, y);
  
    //process all healthy
    el_host = list_first(wf->healthy);
    while (el_host) {
        //try to receive host from other process => polling 
        world_polling_process(neigh_processes, time);
        host = list_element_get_value(el_host);
        el_host = list_element_next(el_host);
        host_prepare_infection_process(host, time);
        host_process(host, time);
    }
}

/**
 *  Process all infected/ill/dead hosts for a field.
 *  The difference after processing is the result
 *  of one time step evaluated.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 *  @param time            current time step
 */
static void world_process_infected_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time)
{
    static Field* wf;
    static Host* host;
    static List_element* el_host;
    wf = get_field_from_matrix_for_coord(x, y);
    
    //process all infected
    el_host = list_first(wf->infected);
    while (el_host) {
        //try to receive host from other process => polling
        world_polling_process(neigh_processes, time);
        host = list_element_get_value(el_host);
        /* muss vor host_process gerufen werden
           damit next nicht am pot. entfernten host
           (nach host_delete) aufgerufen wird */
        el_host = list_element_next(el_host);
        host_process(host, time);
    }
}

/**
 *  Convenience function which wraps both process type functions.
 *
 *  @param x               x_coord of field
 *  @param y               y_coord of field
 *  @param neigh_processes neigh processes -> see polling
 */
static void world_process_both_types_in_field(unsigned x, unsigned y, DynArray* neigh_processes, unsigned time)
{
    world_process_healthy_in_field(x, y, neigh_processes, time);
    world_process_infected_in_field(x, y, neigh_processes, time);
}


static int world_log_timestep_stats(unsigned short curr_time_step)
{
    int rank;
    unsigned current_infected, total_immune, total_died, current_ill, current_healthy, total_infected, total_ill, current_infectious_dead;

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Allreduce(&world.infected_current, &current_infected, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.immune_total, &total_immune, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.died_total, &total_died, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.ill_current, &current_ill, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.healthy_current, &current_healthy, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.infected_total, &total_infected, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.ill_total, &total_ill, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&world.infectious_dead_current, &current_infectious_dead, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);

    if (rank == 0) {
        printf("%u;%u;%u;%u;%u;%u;%u;%u;%u;%u\n", curr_time_step, current_healthy,
               current_infected, current_ill, total_died, current_infectious_dead,
               current_healthy + current_ill + current_infected + total_died,
               total_infected, total_immune, total_ill);
    }

    if (current_infected == 0 && current_ill == 0 && current_infectious_dead == 0) {
        return 0;
    }
    else {
        return 1;
    }
}

/**
 *  Key simulation function. All relevant functions/operations/calculations
 *  for the simulation are contained here and called during main().
 *
 *  @param time_steps number of time steps to calculate
 */
void world_simulate(unsigned short time_steps, unsigned stop_if_finished)
{
    unsigned short time = 0;
    #ifdef DRAW
        world_draw(time);
    #endif

    unsigned x,y;
    unsigned send_count_buf, recv_count_buf;
    
    int size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    
    DynArray neigh_processes = array_create();
   
    //construct list of neighbour processes
    for (int i = -1; i < 2; ++i) {
        for (int j = -1; j < 2; ++j) {
            if (i != 0 || j != 0) {
                int row = rank / world.process_matrix_dimensions.cols;
                int col = rank % world.process_matrix_dimensions.cols;

                row = (row + i + world.process_matrix_dimensions.rows) % 
                  world.process_matrix_dimensions.rows;
                col = (col + j + world.process_matrix_dimensions.cols) %
                  world.process_matrix_dimensions.cols;

                int neighbour_proc = col + row * world.process_matrix_dimensions.cols; 
                if (!array_is_element_contained(&neigh_processes, neighbour_proc)){
//                    printf("my rank %i, neighbour is %i\n", rank, neighbour_proc);
                    array_add(&neigh_processes, neighbour_proc);
                }
            }
        }
    }
    
    world.number_of_neighbours = array_length(&neigh_processes);
    
    if (rank == 0) {
        printf ("**********\n");
        printf ("time;healthy;infected;ill;died;infectious_dead;sum;infected_total;immune_total;ill_total\n");
    }
    
    double starttime, endtime;
    MPI_Barrier(MPI_COMM_WORLD);
    starttime = MPI_Wtime();
    int actionleft = world_log_timestep_stats(time);
    
    while (time < time_steps && (!stop_if_finished || actionleft)) {
        /*––––––––––––––––– move all hosts ––––––––––––––––––––––*/
        //fields at border are processed first
        
        //cols = n = x, rows = m = y
        //x[0 -> max_col_index] y[0]
        //x[0 -> max_col_index] y[max_row_index]
        for (x = 0; x < world.field_matrix->n; ++x) {
            world_move_both_types_in_field(x, 0, &neigh_processes);
            world_move_both_types_in_field(x, (world.field_matrix->m-1), &neigh_processes);
        }
        //x[0] y[1 -> max_row_index-1]
        //x[max_col_index] y[1 -> max_row_index-1]
        for (y = 1; y < world.field_matrix->m -1 ; ++y) {
            world_move_both_types_in_field(0, y, &neigh_processes);
            world_move_both_types_in_field((world.field_matrix->n-1), y, &neigh_processes);
        }
        //process fields inside the border
        for (x = 1; x < (world.field_matrix->n-1); ++x) {
            for (y = 1; (y < world.field_matrix->m-1) ; ++y) {
                world_move_both_types_in_field(x, y, &neigh_processes);
            }
        }
        /*––––––––––––––––– move catch phase ––––––––––––––––––––––*/
        MPI_Allreduce(&world.hosts_send_count, &send_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
        MPI_Allreduce(&world.hosts_received_count, &recv_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
        //if (rank == 0) {
        //   printf("\nbefore catch move: %i %i", send_count_buf, recv_count_buf);
        //}
 
        while (send_count_buf != recv_count_buf) {
            world_polling_move(&neigh_processes);
            MPI_Allreduce(&world.hosts_send_count, &send_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
            MPI_Allreduce(&world.hosts_received_count, &recv_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
            //if (rank == 0) {
            //    printf("\ncatch move: %i %i", send_count_buf, recv_count_buf);
            //}
            
        }
        //if (rank == 0) {
        //   printf("\nafter catch move: %i %i", send_count_buf, recv_count_buf);
        //}
//        printf("\nmoving rank %i: send %i, recv %i\n", rank, world.hosts_send_count, world.hosts_received_count);
        #ifdef DEBUG
            assert(send_count_buf == recv_count_buf);
        #endif
        world.hosts_received_count = 0;
        world.hosts_send_count = 0;
        
        /*–––––––––––– process all hosts ––––––––––––––––––––––––––*/
        //fields at border are processed first
        
        //cols = n = x, rows = m = y
        //x[0 -> max_col_index] y[0]
        //x[0 -> max_col_index] y[max_row_index]
        for (x = 0; x < world.field_matrix->n; ++x) {
            world_process_both_types_in_field(x, 0, &neigh_processes, time);
            world_process_both_types_in_field(x, (world.field_matrix->m-1), &neigh_processes, time);
        }
        //x[0] y[1 -> max_row_index-1]
        //x[max_col_index] y[1 -> max_row_index-1]
        for (y = 1; y < world.field_matrix->m -1; ++y) {
            world_process_both_types_in_field(0, y, &neigh_processes, time);
            world_process_both_types_in_field((world.field_matrix->n-1), y , &neigh_processes, time);
        }
        
        //process fields inside the border
        for (x = 1; x < world.field_matrix->n-1 ; ++x) {
            for (y = 1; y < world.field_matrix->m-1; ++y) {
                world_process_both_types_in_field(x, y, &neigh_processes, time);
            }
        }
        
        /*––––––––––––––––– infect catch phase ––––––––––––––––––––––*/
        MPI_Allreduce(&world.infects_send_count, &send_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
        MPI_Allreduce(&world.infects_received_count, &recv_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);

        //if (rank == 0) {
        //    printf("\nbefore catch infect: %i %i", send_count_buf, recv_count_buf);
        //}

        while (send_count_buf != recv_count_buf) {
            world_polling_process(&neigh_processes, time);
            MPI_Allreduce(&world.infects_send_count, &send_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
            MPI_Allreduce(&world.infects_received_count, &recv_count_buf, 1, MPI_UNSIGNED, MPI_SUM, MPI_COMM_WORLD);
        }
        
        //if (rank == 0) {
        //    printf("\nafter catch infect: %i %i", send_count_buf, recv_count_buf);
        //    printf("\ntime: %u", time);
        //}
//        printf("\nprocessing rank %i: send %i, recv %i\n", rank, world.infects_send_count, world.infects_received_count);
//        printf("\nresidents (ill, infected) rank %i: %i (%i, %i)\n", rank, world.residents, world.ill_total, world.infected_total);
        #ifdef DEBUG
            assert(send_count_buf == recv_count_buf);
        #endif
 
        world.infects_received_count = 0;
        world.infects_send_count = 0;

        time++;
        actionleft = world_log_timestep_stats(time);
        #ifdef DRAW
        if (rank == 0 || rank == 1 || rank == 2 || rank == 3) {
            world_draw(time);
        }
        #endif
    }

    if (rank == 0) {
        printf ("**********\n");
    }

    //print out total runtime
    MPI_Barrier(MPI_COMM_WORLD);
    endtime = MPI_Wtime();
    if (rank == 0) {
        printf("\ntotal time: %f\n", endtime-starttime);
    }

}
